from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class NoSimulatedMetrics(Rule):
    """
    Ensures there are no `simulatedMetricsContainer`s, which were removed with Xcode 9
    """
    def check(self, context):  # type: (XibContext) -> None
        root = context.tree.getroot()
        for container in root.findall('.//simulatedMetricsContainer'):
            context.error(container, 'SimulatedMetricsContainers should be removed (resave with Xcode 9+)')



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class StrictFontNames(Rule):
    """
    Ensures fonts are in the allowed set.

    Example configuration:
    {
      "allowed_fonts": ["ComicSans-Regular", "ComicSans-Bold"],
      "allow_system_fonts": true
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        allowed_fonts = self.config.get('allowed_fonts', [])
        allow_system_fonts = self.config.get('allow_system_fonts', False)

        for element in context.tree.findall(".//fontDescription") + context.tree.findall(".//font"):
            # Skip <font> tags nested in a localization comment
            if element.tag == 'font':
                container = element.parent.parent.parent
                if container.tag == 'attributedString' and container.get('key') == 'userComments':
                    continue

            font_name = element.get('name')
            if font_name is None:
                if not allow_system_fonts:
                    context.error(element, "Use of system fonts is not allowed.")
                continue

            if font_name not in allowed_fonts:
                context.error(element, '"{}" is not one of the allowed fonts.'.format(font_name))



import re

from xiblint.rules import Rule
from xiblint.xibcontext import XibContext
from xiblint.xibutils import (
    get_object_id,
    get_view_user_defined_attr,
    view_accessibility_identifier,
)


class AccessibilityFormat(Rule):
    """
    Checks for incorrect use of Lyft extensions `accessibilityFormat` and `accessibilitySources`.
    """
    def check(self, context):  # type: (XibContext) -> None
        views_with_accessibility_format = {
            element.parent.parent
            for element in context.tree.findall(
                ".//userDefinedRuntimeAttributes/userDefinedRuntimeAttribute[@keyPath='accessibilityFormat']")
        }
        views_with_accessibility_sources = {
            element.parent.parent
            for element in context.tree.findall(".//connections/outletCollection[@property='accessibilitySources']")
        }

        for view in views_with_accessibility_format | views_with_accessibility_sources:
            check_view(context, view)


def check_view(context, view):
    accessibility_source_destination_ids = [
        outlet_connection.get('destination') for outlet_connection
        in view.findall("./connections/outletCollection[@property='accessibilitySources']")
    ]

    accessibility_sources = [
        context.tree.find(".//*[@id='{}']".format(destination_id))
        for destination_id in accessibility_source_destination_ids
    ]
    accessibility_sources_count = len(accessibility_sources)

    #
    # Ensure accessibilityFormat iff accessibilitySources
    #
    accessibility_format = get_view_user_defined_attr(view, 'accessibilityFormat')
    if accessibility_format is None and accessibility_sources_count != 0:
        context.error(view, "View has {} accessibility source(s) but accessibilityFormat is unset.",
                      accessibility_sources_count)
        return

    used_identifiers = set(get_accessibility_identifiers(accessibility_format))
    if used_identifiers:
        check_new_format(context, view, accessibility_format, accessibility_sources, used_identifiers)
    else:
        check_old_format(context, view, accessibility_format, accessibility_sources)


def check_old_format(context, view, accessibility_format, accessibility_sources):
    #
    # Check for strange format types (e.g. %d)
    #
    for m in re.finditer(r'%(.)', accessibility_format):
        format_type = m.group(1)
        if format_type not in ['@', '%']:
            context.error(view, "Unexpected format type %{} in accessibility format '{}'", format_type,
                          accessibility_format)

    #
    # Check for accessibilityFormat that could've been a accessibilityLabel
    #
    expected_sources_count = len(re.findall(r'%@', accessibility_format))
    if expected_sources_count == 0 and '[self]' not in accessibility_format:
        context.error(view, "No format specifiers in '{}'; use `accessibilityLabel` instead", accessibility_format)

    #
    # Check old format (with %@-s).
    #
    if len(accessibility_sources) != expected_sources_count:
        context.error(view, "Format string '{}' has has {} format specifiers, but view has {} sources.",
                      accessibility_format, expected_sources_count, len(accessibility_sources))


def check_new_format(context, view, accessibility_format, accessibility_sources, used_identifiers):
    sources_by_id = get_accessibility_sources_by_id(accessibility_sources)
    source_identifiers = set(sources_by_id.keys())
    missing_identifiers = used_identifiers - source_identifiers
    extra_source_identifiers = source_identifiers - used_identifiers
    for identifier in missing_identifiers:
        context.error(view, "Missing accessibility source for identifier '{}' in accessibility format", identifier)
    for identifier in extra_source_identifiers:
        context.error(view, "Unused accessibility source with identifier '{}' in accessibility format", identifier)
    unidentified_sources = set(accessibility_sources) - set(sources_by_id.values())
    for source in unidentified_sources:
        context.error(view, "Accessibility source '{}' missing accessibility identifier", get_object_id(source))
    if '%@' in accessibility_format:
        context.error(view, "New-style accessibility format contains '%@'")


def get_accessibility_identifiers(accessibility_format):
    for m in re.finditer(r'\[(.+?)\]', accessibility_format):
        identifier = m.group(1)
        if identifier != 'self':
            yield identifier


def get_accessibility_sources_by_id(accessibility_sources):
    """
    Gets a dictionary of accessibility sources keyed by their 'accessibilityFormatIdentifier'.
    """
    accessibility_sources_by_id = {}
    for source in accessibility_sources:
        accessibility_identifier = view_accessibility_identifier(source)
        if accessibility_identifier:
            accessibility_sources_by_id[accessibility_identifier] = source
    return accessibility_sources_by_id



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class NoTraitVariations(Rule):
    """
    Checks for Trait Variations being enabled.
    """
    def check(self, context):  # type: (XibContext) -> None
        root = context.tree.getroot()
        if root.get('useTraitCollections') == 'YES' and root.get('targetRuntime') != 'watchKit':
            context.error(root, "Document has 'Use Trait Variations' enabled")



import inspect
from glob import glob
from os.path import basename, dirname
from importlib import import_module


class Rule(object):
    def __init__(self, config):
        self.config = config

    def check(self, context):
        raise NotImplementedError()


def _class_name_for_file(filename):
    return filename.title().replace("_", "")


def _collect_checkers():
    _rule_checkers = {}
    for filepath in glob(dirname(__file__) + "/*.py"):
        rule_name = basename(filepath)[:-3]
        if rule_name == "__init__":
            continue

        module = import_module('xiblint.rules.' + rule_name)
        class_name = _class_name_for_file(rule_name)
        klass = getattr(module, class_name)
        if inspect.isclass(klass):
            _rule_checkers[rule_name] = klass
        else:
            raise TypeError("Expected function in {}".format(filepath))
    return _rule_checkers


rule_checkers = _collect_checkers()



#!/usr/bin/env python

import argparse
import json
import os
import sys

from xiblint import __version__


def make_epilog_text():
    import xiblint.rules
    description = "Lint rules:\n"

    for rule_name in sorted(xiblint.rules.rule_checkers.keys()):
        checker = xiblint.rules.rule_checkers[rule_name]
        description += "\t{}\n".format(rule_name)
        doc = checker.__doc__
        if doc is not None:
            for line in doc.strip().split("\n"):
                description += " " * 24 + "{}\n".format(line.strip())

    return description


def main():
    from .patch_element_tree import patch_element_tree
    patch_element_tree()

    parser = argparse.ArgumentParser(
        description='.xib / .storyboard linter',
        epilog=make_epilog_text(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("-v", "--version", action="version",
                        version=__version__)
    parser.add_argument("--reporter", choices=("raw", "json"),
                        default="raw",
                        help="custom reporter to use")
    parser.add_argument("paths", nargs=argparse.REMAINDER,
                        help="lint only at the specified paths")
    args = parser.parse_args()

    try:
        from xiblint.config import Config
        config = Config()
    except IOError as ex:
        print('Error: {}\n'.format(ex))
        parser.print_usage()
        sys.exit(1)
    except ValueError as ex:
        print('Error: {}: {}\n'.format(Config.filename, ex))
        parser.print_usage()
        sys.exit(1)

    #
    # Process paths
    #
    errors = []
    include_paths = args.paths or config.include_paths
    for path in include_paths:
        if os.path.isfile(path):
            errors += process_file(path, config)
        elif os.path.isdir(path):
            for root, _, files in os.walk(path):
                for filename in files:
                    if os.path.splitext(filename)[1].lower() in ('.storyboard', '.xib'):
                        file_path = os.path.join(root, filename)
                        errors += process_file(file_path, config)
        else:
            print("Error: Invalid path '{}'".format(path))
            sys.exit(1)

    print_errors(errors, args.reporter)

    sys.exit(1 if errors else 0)


def process_file(file_path, config):
    from xiblint.xibcontext import XibContext
    checkers = config.checkers(file_path)
    context = XibContext(file_path)
    if context.tree:
        for rule_name, klass in checkers.items():
            context.rule_name = rule_name
            rule_config = config.config_for_rule(file_path, rule_name)
            instance = klass(rule_config)
            instance.check(context)
    return context.errors


def print_errors(errors, reporter):
    if reporter == "raw":
        for error_dict in errors:
            error_str = error_dict['error']
            if 'rule' in error_dict:
                error_str += ' [rule: {}]'.format(error_dict['rule'])
            print("{}:{}: error: {}".format(
                error_dict["file"],
                error_dict["line"],
                error_str,
            ))
    elif reporter == "json":
        print(json.dumps(errors))


if __name__ == '__main__':
    main()



from __future__ import unicode_literals

from defusedxml import ElementTree

from .xmlutils import parse_xml
from .xibutils import get_object_id


class XibContext(object):
    def __init__(self, path):
        self.path = path
        self.success = True
        self.errors = []
        try:
            self.tree = parse_xml(path)
        except ElementTree.ParseError as ex:
            self.tree = None
            self.errors.append({
                'file': self.path,
                'line': ex.position[0],
                'error': ex.msg,
            })
        self.rule_name = None

    @staticmethod
    def _get_moniker(view):
        """
        Tries to determine a human-readable string for a view.
        """
        label = view.get('userLabel')
        if label:
            return label
        if view.tag == 'imageView':
            return view.get('image')
        if view.tag == 'button':
            state_normal = view.find("./state[@key='normal']")
            if state_normal is not None:
                return state_normal.get('title') or state_normal.get('image')

        return None

    def error(self, element, message, *args):
        self.success = False
        object_id = get_object_id(element)
        moniker = self._get_moniker(element)

        new_error = "{}{}{}".format(
            "{}: ".format(object_id) if object_id else '',
            "'{}': ".format(moniker) if moniker else '',
            message.format(*args),
        )
        new_error_dict = {
            "file": self.path,
            "line": element.line,
            "error": new_error,
            "rule": self.rule_name,
        }

        self.errors.append(new_error_dict)



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class EnforceRuntimeAttributes(Rule):
    """
    Ensures a runtime attribute is set to one of the allowed values.

    You must specify a module as part of the class name if it is a custom type.

    Example configuration:
    {
      "runtime_attributes": {
        "SomeModule.LegacyButton": {
          "sizeName": ["small", "large"]
        },
        "button": {
          "layer.cornerRadius": [null]
        }
      }
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        runtime_attributes = self.config.get('runtime_attributes', {})

        for full_class_name in runtime_attributes.keys():
            enforced_attributes = runtime_attributes.get(full_class_name)

            # Get system or custom class elements
            full_class_name_split = full_class_name.split(".")
            if len(full_class_name_split) == 1:
                elements = context.tree.findall(".//{}".format(full_class_name))
            elif len(full_class_name_split) == 2:
                elements = context.tree.findall(".//*[@customClass='{}'][@customModule='{}']"
                                                .format(full_class_name_split[1], full_class_name_split[0]))

            for element in elements:
                for attribute_keyPath in enforced_attributes.keys():
                    attribute_allowed_values = enforced_attributes.get(attribute_keyPath)
                    attribute_value = None

                    attribute_list = element.find("./userDefinedRuntimeAttributes")
                    if attribute_list is not None:
                        attribute_element = attribute_list.find("./userDefinedRuntimeAttribute/[@keyPath='{}']"
                                                                .format(attribute_keyPath))
                        if attribute_element is not None:
                            attribute_value = attribute_element.get("value")

                    if attribute_value in attribute_allowed_values:
                        continue

                    options_string = '`, `'.join(map(str, attribute_allowed_values))
                    context.error(element, '`<{}>` runtime attribute `{}` must use `{}` instead of `{}`.'
                                  .format(full_class_name, attribute_keyPath, options_string, attribute_value))



def patch_element_tree():
    # HACK to ensure cElementTree is not loaded -- we need the pure Python implementation
    import sys
    assert 'xml.etree.ElementTree' not in sys.modules
    sys.modules['_elementtree'] = None  # type: ignore



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class NamedColors(Rule):
    """
    Ensures colors are using named colors from an asset catalog.

    Example configuration:
    {
      "allowed_colors": ["CustomRed", "CustomGreen"],
      "allow_system_colors": true,
      "ignore_alpha": true,
      "allow_clear_color": true
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        allowed_colors = self.config.get('allowed_colors', [])
        allow_system_colors = self.config.get('allow_system_colors', False)
        ignore_alpha = self.config.get('ignore_alpha', False)
        allow_clear_color = self.config.get('allow_clear_color', False)

        for element in context.tree.findall(".//color"):
            # Skip <color> tags nested in a localization comment
            container = element.parent.parent.parent
            if container.tag == 'attributedString' and container.get('key') == 'userComments':
                continue

            # Skip <color> tags part of a color or system color definition
            if element.parent.tag == 'namedColor' or element.parent.tag == 'systemColor':
                continue

            # Skip colors with alpha (if configured)
            if ignore_alpha and element.get('alpha') is not None and element.get('alpha') != '1':
                continue

            # Skip clear color (if configured)
            if (
                    allow_clear_color and
                    element.get('alpha') is not None and element.get('white') is not None and
                    element.get('alpha') == '0.0' and element.get('white') == '0.0'
            ):
                continue

            # If `systemColor` or `catalog` is present, it's a named system color
            if (
                    element.get('systemColor') is not None or
                    element.get('cocoaTouchSystemColor') is not None or
                    element.get('catalog') is not None
            ):
                if not allow_system_colors:
                    context.error(element, "Use of named system colors is not allowed. Use a named color instead.")
                continue

            # Require a name
            color_name = element.get('name')
            if color_name is None:
                context.error(element, "Use of custom colors is not allowed. Use a named color instead.")
                continue

            # If allowed_colors is set, verify that color_name is included
            options_string = '`, `'.join(map(str, allowed_colors))
            if allowed_colors and color_name not in allowed_colors:
                context.error(element, '"{}" is not one of the allowed colors: `{}`.'
                              .format(color_name, options_string))



from xml.etree.ElementTree import Element

from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class UnavailableCustomClasses(Rule):
    """
    Ensures a given custom class isn't used and provides a replacement suggestion.

    You must specify a module as part of the class name.

    Example configuration:
    {
      "custom_classes": {
          "SomeModule.LegacyButton": "SomeModule.NewButton"
      }
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        unavailable_classes = self.config.get('custom_classes', {})

        for element in context.tree.findall('.//*[@customClass]'):
            full_name = self._full_class_name(element)
            replacement = unavailable_classes.get(full_name)

            if not replacement:
                continue

            context.error(element, '`{}` is prohibited. Use `{}` instead.'.format(full_name, replacement))

    def _full_class_name(self, element):  # type: (Rule, Element) -> str
        return "{}.{}".format(element.get('customModule'), element.get('customClass'))



from xml.etree.ElementTree import Element

from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class UnavailableSystemClasses(Rule):
    """
    Ensures given system types are subclassed by a set of classes.

    You must specify a module as part of the class name.

    Example configuration:
    {
      "system_classes": {
          "navigationController": ["ModuleName.CustomNavigationController"],
          "button": ["ModuleName.CoolButton", "ModuleName.CoolerButton"]
      }
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        custom_classes = self.config.get('system_classes', {})

        for tag_name in custom_classes.keys():
            for element in context.tree.findall('.//{}'.format(tag_name)):
                tag_name = element.tag
                options = custom_classes.get(tag_name)
                custom_class = element.get('customClass')
                options_string = '`, `'.join(options)

                if not custom_class:
                    context.error(element, '`<{}>` without a custom class is prohibited. Use `{}` instead.'
                                  .format(tag_name, options_string))
                    continue

                full_class_name = self._full_class_name(element)
                if full_class_name in options:
                    continue

                context.error(element, '`<{}>` must use `{}` instead of `{}`.'
                              .format(tag_name, options_string, full_class_name))

    def _full_class_name(self, element):  # type: (Rule, Element) -> str
        return "{}.{}".format(element.get('customModule'), element.get('customClass'))



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext
import glob
import json


class ColorAssets(Rule):
    def __init__(self, config):
        asset_catalog_path = config.get("asset_catalog", None)
        if asset_catalog_path is None:
            raise SystemExit(
                "error: Asset catalog not found. Please configure 'asset_catalog'.",
            )

        self.assets = glob.glob("{}/**/*.colorset".format(asset_catalog_path))

        if not self.assets:
            raise SystemExit(
                "error: Failed to load asset catalog at: '{}'".format(
                    asset_catalog_path,
                ),
            )

        self.colors = {}

    """
    Ensures cached named colors match the named colors from an asset catalog.

    Example configuration:
    {
      "asset_catalog": "Resources/Color.xcassets"
    }
    """

    def check(self, context):  # type: (XibContext) -> None
        for element in context.tree.findall(".//namedColor"):
            name = element.get("name")
            if not name:
                context.error(element, "Named color is missing a name.")
                continue

            expected_color = self._load_color(name)
            if not expected_color:
                context.error(
                    element,
                    "Named color '{}' is missing from asset catalog.".format(name),
                )
                continue

            color_element = element.find("color")
            if color_element is None:
                context.error(
                    element,
                    "Named color '{}' is missing color definition.".format(name),
                )
                continue

            red = color_element.get("red")
            green = color_element.get("green")
            blue = color_element.get("blue")
            alpha = color_element.get("alpha")
            if not all([red, green, blue, alpha]):
                context.error(
                    element, "Named color '{}' has invalid color value.".format(name),
                )
                continue

            color = (float(red), float(green), float(blue), float(alpha))

            if not self._compare_colors(color, expected_color):
                print(color, expected_color)
                context.error(
                    element,
                    "Color value for '{}' does not match asset catalog.".format(name),
                )

    def _load_color(self, name):
        if name in self.colors:
            return self.colors[name]

        for path in self.assets:
            if not path.endswith("{}.colorset".format(name)):
                continue

            with open("{}/Contents.json".format(path)) as json_file:
                data = json.load(json_file)
                color = self._load_components(data["colors"][0]["color"]["components"])
                self.colors[name] = color
                return color

    def _load_components(self, components):
        return (
            self._load_component(components, "red"),
            self._load_component(components, "green"),
            self._load_component(components, "blue"),
            self._load_component(components, "alpha"),
        )

    def _load_component(self, components, key):
        string = components[key]

        if "." in string:
            return float(string)

        return float(string) / 255.0

    def _compare_colors(self, lhs, rhs):
        # Compare with only 4 decimal places
        return all(format(l, ".4f") == format(r, ".4f") for l, r in zip(lhs, rhs))



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext
from xiblint.xibutils import (
    view_is_accessibility_element,
    view_accessibility_label,
    get_view_user_defined_attr,
)


class AccessibilityLabelsForImages(Rule):
    """
    Checks for accessible images with no accessibility label.
    In this case, VoiceOver will announce the image asset's name, which might be unwanted.
    """
    def check(self, context):  # type: (XibContext) -> None
        for image_view in context.tree.findall(".//imageView"):
            if (
                    view_is_accessibility_element(image_view) is True and
                    not view_accessibility_label(image_view) and
                    not get_view_user_defined_attr(image_view, 'accessibilityFormat')
            ):
                context.error(image_view, "Image is accessible but has no accessibility label")



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class AutomationIdentifiersForOutletLabels(Rule):
    """
    Checks for labels with outlets into a view controller that have no accessibility identifiers.
    Labels with outlets might get dynamic text, and therefore should be accessible to UI testing.
    """
    def check(self, context):  # type: (XibContext) -> None
        for outlet in context.tree.findall(".//viewController/connections/outlet"):
            destination = outlet.get('destination')
            label = context.tree.find(".//label[@id='{}']".format(destination))
            if label is None:
                continue
            if label.find('./accessibility[@identifier]') is not None:
                continue
            context.error(
                label,
                "Label with text '{}' has an outlet into the view controller "
                "so it requires an accessibility identifier",
                label.get('text'), label.tag)



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class AutomationIdentifiers(Rule):
    """
    Makes sure that interactive views have accessibility identifiers, to support testing through UI Automation.
    """
    def check(self, context):  # type: (XibContext) -> None
        for tag in ['button', 'textField', 'textView']:
            for view in context.tree.findall(".//{}".format(tag)):
                if view.find('./accessibility[@identifier]') is None:
                    context.error(view, "{} requires an accessibility identifier for UI Automation", tag)



from .xmlutils import element_and_parents


def get_object_id(element):
    for child in element_and_parents(element):
        object_id = child.get('id')
        if object_id is not None:
            return object_id

    return None


def get_view_user_defined_attr(view, key_path):
    attr = view.find("./userDefinedRuntimeAttributes/userDefinedRuntimeAttribute[@keyPath='{}']".format(key_path))
    return attr.get('value') if attr is not None else None


def view_accessibility_label(view):
    config = view.find("accessibility[@key='accessibilityConfiguration']")
    return config.get('label') if config is not None else None


def view_accessibility_identifier(view):
    config = view.find("accessibility[@key='accessibilityConfiguration']")
    return config.get('identifier') if config is not None else None


def view_is_accessibility_element(view):
    is_element = view.find("accessibility[@key='accessibilityConfiguration']/bool[@key='isElement']")
    if is_element is None:
        return False
    return is_element.get('value') == 'YES'



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class StrictFontSizes(Rule):
    """
    Ensures fonts are in the allowed set.

    Example configuration:
    {
      "minimum_size": 13,
      "maximum_size": 30
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        minimum_size = self.config.get('minimum_size', 0)
        maximum_size = self.config.get('maximum_size', 1000)

        for element in context.tree.findall('.//font') + context.tree.findall('.//fontDescription'):
            attribute_name = None

            if element.tag == 'font':
                # Skip <font> tags nested in a localization comment
                container = element.parent.parent.parent
                if container.tag == 'attributedString' and container.get('key') == 'userComments':
                    continue

                attribute_name = 'size'
            else:
                attribute_name = 'pointSize'

            size = element.get(attribute_name)
            if size is None:
                context.error(element, 'Invalid <{}> found. Must have a {}.'.format(element.tag, attribute_name))
                continue

            point_size = int(size)

            if point_size < minimum_size:
                context.error(element, '"{}" is smaller than the allowed minimum size ({})'.format(size, minimum_size))
            elif point_size > maximum_size:
                context.error(element, '"{}" is larger than the allowed maximum size ({})'.format(size, maximum_size))



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class NoViewControllerLinksToOtherBundles(Rule):
    """
    Ensures there are no links to storyboards in different bundles
    """
    def check(self, context):  # type: (XibContext) -> None
        for element in context.tree.findall(".//viewControllerPlaceholder"):
            if element.get("bundleIdentifier"):
                context.error(element, "Linking to a view controller in another bundle")



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class AutolayoutFrames(Rule):
    """
    Checks for ambiguous and misplaced views.
    """
    def check(self, context):  # type: (XibContext) -> None
        for element in context.tree.iter():
            if element.get('ambiguous') == 'YES':
                context.error(element, "View with ambiguous constraints")
            elif element.get('misplaced') == 'YES':
                context.error(element, "Misplaced view")



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext
from xiblint.xibutils import (
    get_view_user_defined_attr,
    view_is_accessibility_element,
)


class AccessibilityLabelsForTextWithPlaceholder(Rule):
    """
    Checks for text fields and text views with a placeholder and no accessibility label.
    A placeholder is not a substitute for an accessibility label, since it's no longer announced
    after the text is edited.
    """
    def check(self, context):  # type: (XibContext) -> None
        for element in context.tree.findall(".//textField") + context.tree.findall(".//textView"):
            placeholder = (element.get('placeholder') if element.tag == 'textField'
                           else get_view_user_defined_attr(element, 'placeholder'))
            if (
                    placeholder is not None and
                    element.find('./accessibility[@label]') is None and
                    view_is_accessibility_element(element) is not False
            ):
                context.error(
                    element,
                    "{} with placeholder text '{}' but no accessibility label",
                    element.tag, element.get('placeholder'))



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext

ALLOWED_DEVICES = [
    'retina4_0',
    'watch38',
]


class SimulatedMetricsRetina40(Rule):
    """
    Ensures simulated metrics are for the iPhone SE or a 38mm watch
    which are currently the smallest display profiles.
    """
    def check(self, context):  # type: (XibContext) -> None
        root = context.tree.getroot()

        # In Xcode 9 this metadata is in a new place
        device = root.find('device')
        if device is not None and device.get('id') not in ALLOWED_DEVICES:
            context.error(
                device,
                'Simulated metrics ("View As:") must be one of: {}'
                .format(', '.join(ALLOWED_DEVICES)))



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext
from xiblint.xibutils import (
    view_is_accessibility_element,
    view_accessibility_label,
    get_view_user_defined_attr,
)


class AccessibilityLabelsForImageButtons(Rule):
    """
    Checks for image buttons with no accessibility label.
    In this case, VoiceOver will announce the image asset's name, which might be unwanted.
    """
    def check(self, context):  # type: (XibContext) -> None
        for button in context.tree.findall(".//button"):
            state_normal = button.find("./state[@key='normal']")
            if (
                    state_normal is None or
                    'title' in state_normal.attrib or
                    view_is_accessibility_element(button) is False or
                    view_accessibility_label(button) or
                    get_view_user_defined_attr(button, 'accessibilityFormat')
            ):
                continue

            if 'image' in state_normal.attrib:
                context.error(
                    button,
                    "Button with image '{}' and no title; "
                    "should either have an accessibility label or 'Accessibility Enabled' unchecked",
                    state_normal.attrib['image'])

            if 'backgroundImage' in state_normal.attrib:
                context.error(
                    button,
                    "Button with background image '{}' and no title "
                    "should either have an accessibility label or 'Accessibility Enabled' unchecked",
                    state_normal.attrib['backgroundImage'])



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class NoAttributedStringColors(Rule):
    """
    Ensures attributed strings don't specify colors.
    """
    def check(self, context):  # type: (XibContext) -> None
        for element in context.tree.findall(".//color"):
            # Skip <color> tags nested in a localization comment
            container = element.parent.parent.parent
            if container.tag == 'attributedString' and container.get('key') == 'userComments':
                continue

            # Skip <color> tags part of a color definition
            if element.parent.tag == 'namedColor':
                continue

            # Require a name
            if element.get('key') == "NSColor":
                context.error(element, "Colors in attributed strings are not allowed. Configure in code instead.")
                continue



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class StrictFonts(Rule):
    """
    Ensures font name and size combinations are in the allowed set. This would generally be used instead
    of strict_font_names and strict_font_sizes.

    Example configuration:
    {
      "allowed_fonts": [
        {
          "name": "ComicSans-Regular",
          "size": 14
        },
        {
          "name": "ComicSans-Bold",
          "size": 14
        }
      ]
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        allowed_fonts = [(d["name"], d["size"]) for d in self.config.get('allowed_fonts', [])]

        for element in context.tree.findall('.//font') + context.tree.findall('.//fontDescription'):
            size_attribute_name = None

            if element.tag == 'font':
                # Skip <font> tags nested in a localization comment
                container = element.parent.parent.parent
                if container.tag == 'attributedString' and container.get('key') == 'userComments':
                    continue

                size_attribute_name = 'size'
            else:
                size_attribute_name = 'pointSize'

            raw_size = element.get(size_attribute_name)
            if raw_size is None:
                context.error(element, 'Invalid <{}> found. Must have a {}.'.format(element.tag, size_attribute_name))
                continue

            size = int(raw_size)

            name = element.get('name')
            if name is None:
                context.error(element, 'Invalid <{}> found. Must have a name.'.format(element.tag))
                continue

            if (name, size) not in allowed_fonts:
                context.error(element, 'Invalid font found {} {}. Please use an allowed font.'.format(name, size))



from xiblint.rules import Rule
from xiblint.xibcontext import XibContext


class EnforceSystemProperties(Rule):
    """
    Ensures a property in a system type is set to one of the allowed values.

    Example configuration:
    {
      "system_properties": {
        "label": {
          "adjustsFontForContentSizeCategory": ["YES"]
        },
        "button": {
          "reversesTitleShadowWhenHighlighted": [null, "NO"]
        },
        "stackView/color": {
          "key": [null]
        }
      }
    }
    """
    def check(self, context):  # type: (XibContext) -> None
        system_properties = self.config.get('system_properties', {})

        for tag_name in system_properties.keys():
            for element in context.tree.findall('.//{}'.format(tag_name)):
                enforced_properties = system_properties.get(tag_name)
                for property_name in enforced_properties.keys():
                    property_allowed_values = enforced_properties.get(property_name)
                    property_value = element.get(property_name)

                    if property_value in property_allowed_values:
                        continue

                    options_string = '`, `'.join(map(str, property_allowed_values))
                    context.error(element, '`<{}>` property `{}` must use `{}` instead of `{}`.'
                                  .format(tag_name, property_name, options_string, property_value))
